//
// draw.h
//
// Drawing functions
//
// Mark Baker, 15/2/99

#ifndef __DRAW_H
#define __DRAW_H

#include <string.h>

//
// A small library of handy drawing functions.
// All functions draw to an offscreen buffer so that they can be used to
// (a) double buffer
// (b) set up an offscreen buffer to then scroll on
//
// To display the buffer use DrawBuffer().
//
// This offscreen buffer is defined by a pointer, THIS MUST BE DEFINED BY
// YOUR PROGRAM. An example is given below.
//

/* example:
#include "draw.h"

// allocate an offscreen buffer for drawing
unsigned long MyOffscreenBuffer[BUFFER_SIZE_LONGS];

// global pointer for drawing commands
unsigned long * OffscreenBuffer;

void mymain (void)
	{
	...
	// give offscreen buffer initial value
	OffscreenBuffer = MyOffscreenBuffer;
	...
	}
*/

// OffscreenBuffer -- pointer to offscreen buffer - all drawing commands use this buffer.
extern unsigned long * OffscreenBuffer;


// Use these defintions to declare a buffer of the correct size.
#define BUFFER_SIZE_LONGS	32
#define BUFFER_SIZE_BYTES	(4 * BUFFER_SIZE_LONGS)


// The supplied font is six pixels high and a variable width (average is 4 pixels)
// Use FontStrlen() to find the width of a string in the drawing font.
#define FONT_HEIGHT 6

// ClearBuffer() clears the current offscreen buffer.
// It is implemented as a memset() call, and so is a macro.
#define ClearBuffer()	memset(OffscreenBuffer, 0, BUFFER_SIZE_BYTES);


// DrawBuffer() draws the current offscreen buffer to the LCD display
// It is implemented as a memcpy() call, and so is a macro.
#define DrawBuffer()	memcpy(BackPlane0, OffscreenBuffer, BUFFER_SIZE_BYTES);


// Font Drawing
//
// NOTE: to draw any fonts you must include FONTV46.S in your project, as this
// defines the font.

// PrintFont() will draw strings from FLASH or RAM without errors. 
// If compiled with FANCY defined, it will wrap at the edge of the screen.
// Otherwise drawing will stop at the edge of the screen.
extern void PrintFont (int x, int y, char * string);

// FontStrlen() returns the size in pixels of a given string (from FLASH or RAM).
// This can then be used to align the string onscreen.
extern int FontStrlen (char * string);


// VerticalLine() draws a vertical line
extern void VerticalLine (int x, int y1, int y2);


// HorizontalLine() draws a horizontal line
extern void HorizontalLine (int x1, int x2, int y);


// FillRectangle() fills a rectangle
extern void FillRectangle (int x, int y, int width, int height);

// ClearRectangle() clears a rectangle
extern void ClearRectangle (int x, int y, int width, int height);

// EorRectangle() performs an Exclusive OR on a rectangular area.
// This is useful for highlighting text when combined with other functions
/* example:
	char * string = "TEST";
	PrintFont(0, 0, string);
	EorRectangle(0, 0, FontStrlen(string), FONT_HEIGHT);
*/
extern void EorRectangle (int x, int y, int width, int height);


// This enumeration defines the possible directions the screen can be
// scrolled in.
enum
	{
	DIRECTION_UP = 0,
	DIRECTION_DOWN,
	DIRECTION_LEFT,
	DIRECTION_RIGHT
	};
	

// ScrollScreen() changes from the current offscreen buffer to a new buffer
// by scrolling. As it writes to OffscreenBuffer, NewScreen should NOT be the current
// OffscreenBuffer.
// PARAMETERS:
// 		direction - should be one of DIRECTION_xxxx defined above.
// 		delay - a counter is used to slow down the scrolling. This time taken will change
//				depending on current processor speed.
//		newScreen - a pointer to the new buffer. NOTE newScreen != OffscreenBuffer as the
//				contents of the OffscreenBuffer are changed.
extern void ScrollScreen (unsigned char direction, long delay, unsigned long * newScreen);


#endif