//
// protocol.c
//

//
// An example IR high-level protocol that sends and receives 128 byte sectors.
// This is designed to transfer large blocks of memory from flash, and can be used to swap
// memory card save files, etc.
//

// If you reuse this code you should change the header which is transmitted initially to establish communication
// (seach below for 'YOU SHOULD CHANGE THIS BIT' ...)

// The sector buffer used by the IR protocol is aucSectorBuffer, this is defined in protocol.h
// For simplicity this is statically allocated as being 0x200 - 0x280.
// You must change where the ARM linker places variables to start at 0x280 for this to work.

// Mark Baker, SCEE, 17/5/1999

//
// The IR transfer functions use the sector buffer as a general send/receive buffer.
// Transfer is in 128 byte blocks, as this is the smallest size that can be written to Flash.
//
// This uses sample code provided by SCEI which sends variable-size packets.
// Packet size is a mulitple of 32 bits (4 bytes).
// IR communication is using 4MHz clock speed and 'Remote Control' mode (Waveform).
//
// The communications protocol is as follows:
//  () indicates a note - see below.
//  Each line in the diagram represents a 'unit' of time.
//
//         SENDER                          RECEIVER
//     (1) Sends Header                 Waits for Header
//  +->(2) Delays                   (3) Decides to accept/refuse card
//  |  (4) Receives Response        (5) Sends Response
//  |      ...                          Delays  <---------------------------------------+
//  |      Chooses next sector		    ...                                             |
//  |      Copies sector to RAM         ...                                             |
//  |      Sends sector                 Receives sector       OR  (6) Timeout occurs    |
//  +----- Repeats until all            Stores sector                 Sends Retry       | 
//          sectors sent and            Sends Accept                  Repeats...        | 
//          accepted by Receiver        Repeats until all received     |                | 
//                                       |                             |                | 
//                                       +-----------------------------+----------------+ 
//
//
// Notes:
//
// (1) - Header is as follows:
//       The first 12 bytes are a 'magic' word - "POCKET VCARD"
//       The next 2 bytes are the version number.
//       The final 2 bytes are the size, in sectors, of the card sent  (low:high ordering)
//
// (2) - Delays
//       The application can either busy-wait at 4MHz or switch to a lower power mode.
//       If LOW_POWER_WAIT is defined (see below) the lower power mode will be used.
//       This does not affect the performance of the transfer.
//
// (3) - RECEIVER Decides to accept/refuse card
//       The receiver may refuse the card is it does not have sufficient space in Flash to store it.
//       
// (4) - SENDER Receives Response
//       If a timeout occurs, or the user presses a button, the sending routine finishes.
//       (The receiver will eventually timeout and quit)
//
// (5) - RECEIVER Sends Response
//       The possible responses are:
//            Accept sector/card        "ACCEPTx"
//            Refuse sector/card        "REFUSE_"
//            Retry  sector             "RETRY_x"
//
//        'x' is the sector number which is being Accepted/Retried. 
//        NB this is the logical sector number for the card being sent, not the sector in Flash.
//
// (6) - RECEIVER Timeout occurs
//       If too many timeouts have occurred, the receiver will give up.
//       The RECEIVER timeout is set to be half that of the SENDER, to 
//       allow the SENDER to try again.
//

// standard C functions
#include <stdlib.h>
#include <string.h>

// PDA hardware access
#include "pdaregs.h"
#include "pdakernel.h"

// Flash reading utils
#include "flash2ram.h"
#include "strutil.h"

// key reading functions
#include "keys.h"

// drawing functions
#include "draw.h"

// low-level IR functions
#include "../if/if.h"

#include "protocol.h"

// if LOW_POWER_WAIT is defined, will switch to a lower clockspeed during pauses in IR communication
#define LOW_POWER_WAIT

//
// definitions
//

// useful macro to get memory address in flash from relative sector number
#define FLASH_START_ADDRESS ((unsigned char *)(0x02000000))
#define SECTOR_SIZE_BYTES 128
#define MEM_ADDRESS_FOR_SECTOR_NUMBER(secNum)	\
	(unsigned char *)(FLASH_START_ADDRESS + ((secNum) * SECTOR_SIZE_BYTES))


// Version of protocol - up to you whether you want to change this or not.
// default behaviour is to reject communication from later version senders,
// but be backwardly compatable.
#define VERSION_LOWBYTE  ((unsigned char)0x12)
#define VERSION_HIGHBYTE ((unsigned char)0x01)


// 
// YOU SHOULD CHANGE THIS BIT
//
// This is the header which is sent to establish communications.
// To prevent confusion between different programs, please change
// The following string to your own choice (it must remain 12 characters long)
//
#define HEADER_SIZE_BYTES 16
#define HEADER_MAGIC_SIZE 12
const char strIrHeaderMagic [] = "SMPL_IR_CODE";

#define RESPONSE_SIZE_BYTES 8
#define RESPONSE_TEXT_SIZE 7
const char strResponseAccept [] = "ACCEPT_";
const char strResponseRefuse [] = "REFUSE_";
const char strResponseRetry [] =  "RETRY__";

#define BYTES_SENT  128
#define BITS_SENT	(BYTES_SENT * 8)
#define REPEAT_COUNT 1
#define FRAME_LENGTH 0
#define TIMEOUT_VALUE 10000
// wait for 4 seconds for guide pulse (each pulse is 0.4ms)

#define IR_CLOCK_SPEED SYSCLOCK_FREQ_4MHZ
#define IR_TIMER 1
#define IR_TIMER_VALUE 0x7a12
#define IR_TIMER_CLKDIV 0

#ifdef LOW_POWER_WAIT
  // timings at lower clock speed
  #define WAIT_CLOCK_SPEED	SYSCLOCK_FREQ_62KHZ
  #define ADJUSTED_WAIT_VALUE(v)			((v) >> (IR_CLOCK_SPEED - WAIT_CLOCK_SPEED))

  #define IR_DELAY_RECEIVE_WAIT_START 		ADJUSTED_WAIT_VALUE( 60000)
  #define IR_DELAY_SEND_WAIT 				ADJUSTED_WAIT_VALUE(550000)
  #define IR_DELAY_RECEIVE_WAIT 			ADJUSTED_WAIT_VALUE(270000)
  #define IR_DELAY_SEND_RETRY_EXTRA_WAIT 	ADJUSTED_WAIT_VALUE(100000)
#else
  // timings at 4MHz clock speed 
  #define IR_DELAY_RECEIVE_WAIT_START	 	 60000
  #define IR_DELAY_SEND_WAIT 				550000
  #define IR_DELAY_RECEIVE_WAIT 			270000
  #define IR_DELAY_SEND_RETRY_EXTRA_WAIT 	100000
#endif

#define IF_STATUS_QUIT	0x100
#define IF_STANDBY_BIT 0x02


//
// funciton prototypes
//
__inline void vIrStandby (void);
volatile void vIrDelay (unsigned long ulDelay);
unsigned long ulWaitForIrFlags (int iFlags);
void vWriteSectorBuffer (unsigned short usSectorNumber);

//
// functions
//

// vWriteSectorBuffer()
// safely writes sector buffer to flash, even if we're at the wrong
// clockspeed
void vWriteSectorBuffer (unsigned short usSectorNumber)
	{
	// store current processor speed
	int iOldClockFreq = (int)(*PMFrequency & 0xf);
	
	// switch to 4MHz clock speed for FLASH write
	setSystemClockFrequency(SYSCLOCK_FREQ_4MHZ);

	// write block to flash
	writeFlashRelative(usSectorNumber, (unsigned long *)aucSectorBuffer);

	// return to old processor speed
	setSystemClockFrequency(iOldClockFreq);
	}


// vIrStandby()
// switch the IR to be in power-saving standby mode
__inline void vIrStandby (void)
	{
	*IFStaticControl = IF_STANDBY_BIT;
	}
	
	
// vIrDelay()
// utility function to provide a little delay in IR communications
volatile void vIrDelay (unsigned long ulDelay)
	{
	unsigned long counter = ulDelay;

	#ifdef LOW_POWER_WAIT
	// switch to slower clock speed for wait to conserve power
	setSystemClockFrequency(WAIT_CLOCK_SPEED);
	#endif
	
	// simple busy-wait loop
	while (counter--)
		{}
		
	#ifdef LOW_POWER_WAIT
	// switch to normal clock speed after waiting
	setSystemClockFrequency(IR_CLOCK_SPEED);
	#endif	
	}
	
	
// ulWaitForIrFlags()
// utility function to wait for an IR status flag to be set - 
// useful for waiting for send/receive competion
// also allows key breakout with flag set
unsigned long ulWaitForIrFlags (int iFlags)
	{
	unsigned long status;

	startkeys();
	
	do
		{
		status = IF_RMGetStatus();

		if (keys & KEYEnter)
			{
			status |= IF_STATUS_QUIT;
			break;
			}
		}
	while (!(status & iFlags));

	stopkeys();
	
	return status;
	}


// iIrSendSectors()
// Sends a list of sectors over IR.
// Supply a list of sector numbers to send & number of sectors
//  -- these are 'relative' sector numbers, with 0 being the first sector in the program's blocks.
//     63 being the last sector of the first block, etc..
//     e.g. They must be suitable use use with the kernel 'Write Flash Relative' function
//
// Returns IR_PROTOCOL_FAIL or IR_PROTOCOL_SUCCEED

int iIrSendSectors (int * aiSectorsToSend, int iNumberOfSectorsToSend)
	{
	unsigned long status;
	int iOldClockspeed = (int)*PMFrequency & 0xff;
	int iCurrentSector;
	int iReturnValue = IR_PROTOCOL_SUCCEED;
	char strTemp[10];

	// use sector buffer as receive buffer too.
	unsigned char * ppucBufferTable [] = { aucSectorBuffer };
	
	// set up IR
	IF_RMInit(IR_CLOCK_SPEED);
	TIMER_Init(IR_TIMER, IR_TIMER_VALUE, IR_TIMER_CLKDIV);
	
	// send a header sector saying how many sectors there are
	// magic word
	fstrcpy((char *)aucSectorBuffer, strIrHeaderMagic);
	
	// version number
	aucSectorBuffer[HEADER_MAGIC_SIZE] = VERSION_LOWBYTE;
	aucSectorBuffer[HEADER_MAGIC_SIZE + 1] = VERSION_HIGHBYTE;
	
	// number of sectors used
	aucSectorBuffer[HEADER_MAGIC_SIZE + 2] = (unsigned char)(iNumberOfSectorsToSend & 0xff);
	aucSectorBuffer[HEADER_MAGIC_SIZE + 3] = (unsigned char)((iNumberOfSectorsToSend >> 8) & 0xff);
			
	// send the header
	PrintFont(0, 6, "HEADER");
	DrawBuffer();
	IF_RMSendStart(HEADER_SIZE_BYTES * 8, aucSectorBuffer, REPEAT_COUNT, FRAME_LENGTH);
	
	iCurrentSector = -1;
	
	// rather naughty goto to jump into main sending loop 
	//  - I know, I'll go to structured programming hell for this... sorry...
	goto waitForTransmitCompletion;
		
	// loop until all sectors have been confirmed
	while (iCurrentSector < iNumberOfSectorsToSend)
		{
		// delay for a bit here to allow receiver to get ready
		vIrDelay(IR_DELAY_SEND_WAIT);

		// send current sector
		// copy to SRAM
		vCopyWordsFromFlash((unsigned short *)aucSectorBuffer, 
			(const unsigned short *)MEM_ADDRESS_FOR_SECTOR_NUMBER(aiSectorsToSend[iCurrentSector]),
			SECTOR_SIZE_BYTES / 2);
			
		// send over IR
		IF_RMSendStart(BITS_SENT, aucSectorBuffer, REPEAT_COUNT, FRAME_LENGTH);		

		FormatNumber(strTemp, iCurrentSector);
		PrintFont(0, 24, "SEND     ");
		PrintFont(20, 24, strTemp);
		DrawBuffer();

	waitForTransmitCompletion:		
		// wait until xmit complete bit set
		ulWaitForIrFlags(IF_STATUS_TRANSMIT_COMPLETE);

	waitForReceiverResponse:
		// start receiving response 
		IF_RMReceiveStart(RESPONSE_SIZE_BYTES * 8, ppucBufferTable, REPEAT_COUNT, TIMEOUT_VALUE);

		PrintFont(0, 6, "RESP?   ");
		DrawBuffer();
	
		// wait until receive complete bit set (or TO or keypress)
		status = ulWaitForIrFlags(IF_STATUS_RECEIVE_COMPLETE | IF_STATUS_RECEIVE_TIMEOUT);	
		
		// finished with IR for a short time - turn it off to conserve power
		vIrStandby();
					
		// user quit
		if (status & IF_STATUS_QUIT)
			{
			iReturnValue = IR_PROTOCOL_ABANDONED;
			break;
			}

		// timeout or refse - stop, return IR_PROTOCOL_FAIL
		if ((status & IF_STATUS_RECEIVE_TIMEOUT) ||
			(iCompareStringToFlash((char *)aucSectorBuffer, strResponseRefuse, RESPONSE_TEXT_SIZE) == 1))
			// OR timeout...
			{
			PrintFont(0, 6, "REFUSED  ");
			DrawBuffer();
			
			iReturnValue = IR_PROTOCOL_FAIL;
			break;
			}
		
		// retry - continue
		if (iCompareStringToFlash((char *)aucSectorBuffer, strResponseRetry, RESPONSE_TEXT_SIZE) == 1)
			{
			int iRetrySector = (int)aucSectorBuffer[RESPONSE_TEXT_SIZE];

			PrintFont(0, 6, "RETRY   ");
			DrawBuffer();
			
			// if we get a retry for the next sector, we know they got the one we just sent
			// also if we get a retry for a far sector, it's all gone wrong so quit
			
			if (iRetrySector <= (iCurrentSector + 1))
				{
				// resend this sector again (or send the next one)
				iCurrentSector = iRetrySector;			
				vIrDelay(IR_DELAY_SEND_RETRY_EXTRA_WAIT);
				continue;
				}
			else // iRetrySector > (iCurrentSector + 1)
				{
				// that shouldn't happen! quit
				iReturnValue = IR_PROTOCOL_FAIL;
				break;
				}
			}
			
		// accept - next sector
		if (iCompareStringToFlash((char *)aucSectorBuffer, strResponseAccept, RESPONSE_TEXT_SIZE) == 1)
			{
			PrintFont(0, 6, "ACCEPT   ");
			DrawBuffer();
			
			// check for header acceptance
			if (aucSectorBuffer[RESPONSE_TEXT_SIZE] == 0xff)
				{
				iCurrentSector = 0;
				}
			else
				{
				// should send the next sector after the one that was just accepted			
				// this allows Accept to indicate when a sector is lost
				iCurrentSector = ((int)aucSectorBuffer[RESPONSE_TEXT_SIZE]) + 1;
				}
			continue;
			}
			
		// otherwise - what the hell did I get?? try to get something else
		PrintFont(0, 12, "SCHWA??  ");
		aucSectorBuffer[RESPONSE_SIZE_BYTES] = 0;
		PrintFont(0, 18, (char *)aucSectorBuffer);
		DrawBuffer();
		
		goto waitForReceiverResponse;
		}	
	
	// reset clock to normal
	setSystemClockFrequency(iOldClockspeed);
	
	return iReturnValue;
	}


// iIrReceiveSectors()
// Receives a list of sectors over IR
// Supply a list of sector number to write to & a maximum number of sectors to receieve.
// The number of sectors actually received will be written to *piNumberOfSectorsReceived
//
// Returns IR_PROTOCOL_FAIL, IR_PROTOCOL_SUCCEED or IR_PROTOCOL_ABANDONED
// Stops if user presses enter (returns IR_PROTOCOL_ABANDONED)

int iIrReceiveSectors (int * aiSectorsToRecieve, int iMaximumNumberOfSectorsToReceive, 
	int * piNumberOfSectorsReceived)
	{
	unsigned long status;
	int iOldClockspeed = (int)*PMFrequency & 0xff;
	int iNumberOfSectorsToReceive;
	unsigned char ucAccept;
	int iCurrentSectorNo;
	int iReturnValue = IR_PROTOCOL_SUCCEED;
	char strTemp[10];
	int iRetryCounter;
			
	// use sector buffer as receive buffer too.
	unsigned char * ppucBufferTable [] = { aucSectorBuffer };
	
	// set number of sectors received to be 0 so far
	*piNumberOfSectorsReceived = 0;
	
	// check for free space
	if (iMaximumNumberOfSectorsToReceive <= 0)
		{
		return IR_PROTOCOL_FAIL;
		}
	
	// set up IR
	IF_RMInit(IR_CLOCK_SPEED);
	TIMER_Init(IR_TIMER, IR_TIMER_VALUE, IR_TIMER_CLKDIV);

	// receive header
	PrintFont(0, 6, "HEADER?");
	DrawBuffer();
	iRetryCounter = 2;

	do
		{
		// receive
		IF_RMReceiveStart(HEADER_SIZE_BYTES * 8, ppucBufferTable, REPEAT_COUNT, TIMEOUT_VALUE);

		// wait until receive complete bit set
		status = ulWaitForIrFlags(IF_STATUS_RECEIVE_COMPLETE | IF_STATUS_RECEIVE_TIMEOUT);

		// OR timeout and quit
		if (keys & KEYEnter)
			{
			status |= IF_STATUS_QUIT;
			break;
			}
		}
	while ((status & IF_STATUS_RECEIVE_TIMEOUT) && (iRetryCounter-- > 0));

	// finished with IR for a short time - turn it off to conserve power
	vIrStandby();

	if (status & IF_STATUS_QUIT)
		{
		PrintFont(0, 12, "QUIT!");
		setSystemClockFrequency(iOldClockspeed);
		return IR_PROTOCOL_ABANDONED;
		}

	// see if we want to accept these sectors
	ucAccept = 0;

	// number of sectors used
	iNumberOfSectorsToReceive = ((int)aucSectorBuffer[HEADER_MAGIC_SIZE + 2]) | 
						(((int)aucSectorBuffer[HEADER_MAGIC_SIZE + 3]) << 8);
	
	// check magic word
	if (iCompareStringToFlash((char *)aucSectorBuffer, strIrHeaderMagic, HEADER_MAGIC_SIZE) == 1)
		{
		// magic word OK
		unsigned char ucVersion0 = aucSectorBuffer[HEADER_MAGIC_SIZE];
		unsigned char ucVersion1 = aucSectorBuffer[HEADER_MAGIC_SIZE + 1];

		// check version
		if ((VERSION_HIGHBYTE >= ucVersion1) && (VERSION_LOWBYTE >= ucVersion0))
			{
			// version number OK
			
			// check space remaining
			if (iMaximumNumberOfSectorsToReceive >= iNumberOfSectorsToReceive)
				{
				// accept
				ucAccept = 1;				
				}
			}		
		}
		
	// wait for a little bit to give sender a chance to receive
	vIrDelay(IR_DELAY_RECEIVE_WAIT_START);
	
	if (ucAccept)	
		{
		PrintFont(0, 6, "ACCEPT   ");
		DrawBuffer();

		// send accept message
		fstrcpy((char *)aucSectorBuffer, strResponseAccept);
		aucSectorBuffer[RESPONSE_TEXT_SIZE] = 0xff;
		IF_RMSendStart(RESPONSE_SIZE_BYTES * 8, aucSectorBuffer, REPEAT_COUNT, FRAME_LENGTH);

		// clear retry counter
		iRetryCounter = 0;

		// wait for message to send
		ulWaitForIrFlags(IF_STATUS_TRANSMIT_COMPLETE);

		// for each sector:
		iCurrentSectorNo = 0;
		while (iCurrentSectorNo < iNumberOfSectorsToReceive)
			{
			// receive sector
			IF_RMReceiveStart(BITS_SENT, ppucBufferTable, REPEAT_COUNT, TIMEOUT_VALUE/2);

			FormatNumber(strTemp, iCurrentSectorNo);
			PrintFont(0, 24, "RECV     ");
			PrintFont(20, 24, strTemp);
			DrawBuffer();
		
			// wait until receive complete bit set
			status = ulWaitForIrFlags(IF_STATUS_RECEIVE_COMPLETE | IF_STATUS_RECEIVE_TIMEOUT);
		
			// finished with IR for a short time - turn it off to conserve power
			vIrStandby();

			// user quit - just quit
			if (status & IF_STATUS_QUIT)
				{
				iReturnValue = IR_PROTOCOL_ABANDONED;
				break;
				}

			// OR timeout - if over limimt, send Refuse then quit
			//              if not over limit, send Retry and loop with same sector
			if (status & IF_STATUS_RECEIVE_TIMEOUT)
				{
				if (++iRetryCounter > 2)
					{
					PrintFont(0, 24, "TIME ");
					iReturnValue = IR_PROTOCOL_FAIL;
					break;
					}
				else
					{
					PrintFont(0, 6, "RETRY   ");

					// send retry
					fstrcpy((char *)aucSectorBuffer, strResponseRetry);

					// include sector no in response	
					aucSectorBuffer[RESPONSE_TEXT_SIZE] = (unsigned char)iCurrentSectorNo;
					}
				}
			else
				{
				// get the address to write this sector to
				unsigned short usWriteSectorAddress = aiSectorsToRecieve[iCurrentSectorNo];
				
				//    received - save sector, send Accept and loop for next sector
				PrintFont(0, 6, "ACCEPT   ");
				PrintFont(0, 24, "OK   ");
				
				// clear retry counter
				iRetryCounter = 0;
										
				// write the buffer
				vWriteSectorBuffer(usWriteSectorAddress);
				
				// record number of sectors received
				*piNumberOfSectorsReceived = iCurrentSectorNo + 1;
				
				FormatNumber(strTemp, usWriteSectorAddress);
				PrintFont(0, 12, "SAVE     ");
				PrintFont(20, 12, strTemp);
				
				// prepare response
				// set response to be Accept
				fstrcpy((char *)aucSectorBuffer, strResponseAccept);

				// include sector no in response	
				aucSectorBuffer[RESPONSE_TEXT_SIZE] = (unsigned char)iCurrentSectorNo;
				
				// next sector
				iCurrentSectorNo++;													
				}		

			DrawBuffer();
						
			// wait for a little bit to give sender a chance to start receiving
			vIrDelay(IR_DELAY_RECEIVE_WAIT);

			// send response, whatever it is
			IF_RMSendStart(RESPONSE_SIZE_BYTES * 8, aucSectorBuffer, REPEAT_COUNT, FRAME_LENGTH);

			// wait for message to send
			ulWaitForIrFlags(IF_STATUS_TRANSMIT_COMPLETE);
			}			
		}
	else 
		{
		// allocate sectors, send refuse message
		PrintFont(0, 6, "REFUSE   ");
		DrawBuffer();

		fstrcpy((char *)aucSectorBuffer, strResponseRefuse);
		IF_RMSendStart(RESPONSE_SIZE_BYTES * 8, aucSectorBuffer, REPEAT_COUNT, FRAME_LENGTH);

		// wait for message to send
		ulWaitForIrFlags(IF_STATUS_TRANSMIT_COMPLETE);
		
		// quit with error
		iReturnValue = IR_PROTOCOL_FAIL;
		}
	
	// reset clock to normal
	setSystemClockFrequency(iOldClockspeed);
	
	return iReturnValue;
	}
	